;;;;
;;;; explicit.lisp
;;;;
;;;; George Ferguson, ferguson@cs.rochester.edu,  5 Sep 2002
;;;; Time-stamp: <Thu Sep  5 16:31:16 EDT 2002 ferguson>
;;;;
;;;; Takes Lisp forms created from Jerry Hobbs' notation by the
;;;; hobbs2lisp.pl script and adds an outermost ``forall'' form if
;;;; needed to make explicit any universally quantified variables that
;;;; are left implicitly quantified in his notation.
;;;;
;;;; It's just easier to do this in Lisp than either Java or Perl...
;;;;

(defun lisp2explicit ()
  "Translate lisp forms read on *standard-input* by adding any implicit
universally quantified variables, and print them to *standard-output*."
  (format t "~%;; lisp2explicit: ~A~%~%" (timestamp))
  (loop with form
      while (setq form (read *standard-input* nil nil))
      do (let ((tag (second form))
	       (expr (third form)))
	   (format t "~S~%~%" (list 'defaxiom tag
				    (add-implicit-quantifiers expr))))))

(defun add-implicit-quantifiers (expr)
  "Adds explicit forall quantifiers for implicitly universally quantified
variables in EXPR."
  (let ((unbound (reverse (second (gather expr (list nil nil))))))
    (if (null unbound)
	expr
      (list 'forall unbound expr))))

(defun gather (expr vars)
  "Gathers variables from EXPR into VARS, where the first element of
VARS is the list of bound variables in EXPR so far and the second
element of VARS is the list of unbound variables in EXPR so far. Returns
the updated VARS."
  (cond
   ((null expr)
    vars)
   ((listp expr)
    (case (car expr)
      ((forall exists exists!)
       (gather (third expr) (list (append (first vars) (second expr))
				  (second vars))))
      (t
       (loop for x in (cdr expr)
	   do (gather x vars))
       vars)))
   ((and (symbolp expr)
	 (not (char= (char (string expr) 0) #\*)))
    (let ((bound (first vars))
	  (unbound (second vars)))
      (when (not (member expr bound))
	(pushnew expr (second vars))))
    vars)))

(defun timestamp ()
  "Returns an timestamp in standard unix ctime() format."
  (flet ((downame (n) (elt '("Mon" "Tue" "Wed" "Thu" "Fri" "Sat" "Sun") n))
	 (monthname (n) (elt '("" "Jan" "Feb" "Mar" "Apr" "May" "Jun"
			       "Jul" "Aug" "Sep" "Oct" "Nov" "Dec") n))
	 (tzname (n dstp) (case n
			   (5 (if dstp "EDT" "EST"))
			   (6 (if dstp "CDT" "CST"))
			   (7 (if dstp "MDT" "MST"))
			   (8 (if dstp "PDT" "PST"))
			   (t (format nil "GMT~@D~@[DST~]" n dstp)))))
    (multiple-value-bind (sec min hour date month year dow dstp tz)
	(get-decoded-time)
      (format nil "~3A ~3A ~2D ~2D:~2,'0D:~2,'0D ~A ~4D"
	      (downame dow) (monthname month) date hour min sec
	      (tzname tz dstp) year))))
